use std::ops::Deref;

use chrono::Utc;
use serde::{Deserialize, Deserializer, Serialize};

// historic field
#[derive(Deserialize, Serialize, Default, Debug, PartialEq, Clone)]
pub struct Historic<T> {
    pub values: Vec<(chrono::DateTime<Utc>, T)>,
}

impl<T: Clone + Default> Historic<T> {
    pub fn alter_or_default<F: Fn(&mut T)>(&mut self, f: F) {
        let mut c = if let Some(c) = self.current().cloned() {
            c
        } else {
            T::default()
        };
        f(&mut c);
        self.add(c);
    }
}

impl<T: Clone> Deref for Historic<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        self.current().unwrap()
    }
}

impl<T: Clone> Historic<T> {
    pub fn new() -> Self {
        Self { values: vec![] }
    }

    pub fn current(&self) -> Option<&T> {
        let x = self.values.iter().max_by_key(|(dt, _)| *dt)?;

        Some(&x.1)
    }

    pub fn alter<F: Fn(&mut T)>(&mut self, f: F) {
        let mut c = self.current().unwrap().clone();
        f(&mut c);
        self.add(c);
    }

    pub fn add(&mut self, value: T) {
        self.values.push((chrono::Utc::now(), value));
    }
}

pub trait Historicable: Sized {
    fn historic(self) -> Historic<Self> {
        Historic {
            values: vec![(chrono::Utc::now(), self)],
        }
    }
}

impl Historicable for String {}

/// a unique `Vec<T>`
pub struct UniVec<T>(Vec<T>);

impl<T: PartialEq> UniVec<T> {
    pub fn push(&mut self, val: T) -> bool {
        if self.0.iter().any(|x| *x == val) {
            return false;
        }
        self.0.push(val);
        true
    }
}

impl<T: Serialize> Serialize for UniVec<T> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        Vec::<T>::serialize(&self.0, serializer)
    }
}

impl<'de, T: Deserialize<'de>> Deserialize<'de> for UniVec<T> {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let s = Vec::deserialize(deserializer)?;
        Ok(Self(s))
    }
}

impl<T> Deref for UniVec<T> {
    type Target = Vec<T>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}
