use serde::{Deserialize, Deserializer, Serialize};
use std::fmt::Display;

#[derive(Debug, Clone, PartialEq)]
/// Database IDs
pub enum Id {
    /// String ID
    String(String),
    /// ULID
    ULID(ulid::Ulid),
    /// UUID
    UUID(uuid::Uuid),
}

impl<'de> Deserialize<'de> for Id {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let s = String::deserialize(deserializer)?;
        match s.parse::<ulid::Ulid>() {
            Ok(ulid) => Ok(Id::ULID(ulid)),
            Err(_) => match s.parse::<uuid::Uuid>() {
                Ok(uuid) => Ok(Id::UUID(uuid)),
                Err(_) => Ok(Id::String(s)),
            },
        }
    }
}

impl Serialize for Id {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        match self {
            Id::String(str) => str.serialize(serializer),
            Id::ULID(ulid) => ulid.serialize(serializer),
            Id::UUID(uuid) => uuid.serialize(serializer),
        }
    }
}

impl Id {
    /// Generate a new random ULID based ID
    pub fn new_ulid() -> Self {
        Self::ULID(ulid::Ulid::new())
    }
}

impl Display for Id {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Id::String(s) => s.clone(),
                Id::ULID(ulid) => ulid.to_string(),
                Id::UUID(uuid) => uuid.to_string(),
            }
        )
    }
}
