use arc_swap::ArcSwap;
use parking_lot::{RwLock, RwLockReadGuard};
use std::{
    ops::{Deref, DerefMut},
    sync::Arc,
};

use dashmap::DashMap;
use model::file::File;
use relation::{dereference, IdRef};
use serde::{Deserialize, Serialize};
use store::Store;

use crate::Identifiable;

pub mod field;
pub mod id;
pub mod model;
pub mod relation;
pub mod store;

/// really simple query function. it iterates over every object and matches against `predicate`.
///
/// # Examples
///
/// ```ignore
/// let queried = query_t(|x: &MyType| {
///     x.id.to_string().starts_with("something")
/// });
/// ```
pub fn query_t<
    T: Identifiable + Serialize + Send + Sync + for<'a> Deserialize<'a> + 'static,
    F: Fn(&T) -> bool,
>(
    predicate: F,
    db: &Database,
) -> Vec<Model<T>> {
    let collection = T::model_id();
    let all = Store::get_ids(&collection, &*db.storage);
    let all: Vec<IdRef<T>> = all
        .into_iter()
        .map(|x| IdRef::new(format!("{collection}::{x}")))
        .collect();
    let all = dereference(&all, db);
    all.into_iter()
        .filter(|x| {
            let r = x.read();
            let rf = r.deref();
            (predicate)(rf)
        })
        .collect()
}

/// Database Instance
pub struct Database {
    /// Underlying storage layer
    pub storage: Arc<dyn vfs::FileSystem>,
    /// Write models with schema
    pub named: bool,
    /// Use memory caching
    pub cached: bool,
    /// Memory Cache
    records: Arc<DashMap<String, DashMap<String, Box<dyn ModelObj>>>>,
    watcher: Arc<DashMap<String, Vec<crossbeam::channel::Sender<Box<dyn ModelObj>>>>>,
}

pub trait ModelObj: Send + Sync {
    fn as_any(&self) -> &dyn std::any::Any;
}

pub struct ModelWatcher<T> {
    filter: Box<dyn Fn(&T) -> bool + 'static + Send + Sync>,
    apply: Box<dyn Fn(&T) + 'static + Send + Sync>,
    recv: crossbeam::channel::Receiver<Box<dyn ModelObj>>,
}

impl<T: 'static + Identifiable> ModelWatcher<T> {
    pub fn process(&self) {
        if let Ok(val) = self.recv.recv() {
            self.handle(val)
        }
    }

    pub fn recv(&self) -> Model<T> {
        let res = self.recv.recv().unwrap();
        res.as_any().downcast_ref::<Model<T>>().unwrap().clone()
    }

    pub fn handle(&self, val: Box<dyn ModelObj>) {
        log::trace!("Received watch event");
        let model: &Model<T> = val.as_any().downcast_ref().unwrap();
        if (self.filter)(&model.read()) {
            log::info!("Watcher processing {}", model.full_id());
            (self.apply)(&model.read());
        }
    }

    pub fn try_process(&self) {
        if let Ok(val) = self.recv.recv_timeout(std::time::Duration::from_secs(1)) {
            self.handle(val);
        }
    }
}

impl Clone for Database {
    fn clone(&self) -> Self {
        Self {
            storage: Arc::clone(&self.storage),
            named: self.named.clone(),
            cached: self.cached.clone(),
            records: Arc::clone(&self.records),
            watcher: Arc::clone(&self.watcher),
        }
    }
}

impl Database {
    /// Create a new in memory database
    pub fn in_memory() -> Self {
        Self::ensure_init(Self {
            storage: Arc::new(vfs::MemoryFS::new()),
            cached: false,
            named: false,
            records: Arc::new(DashMap::new()),
            watcher: Arc::new(DashMap::new()),
        })
    }

    pub fn watch<T, F, G>(&self, filter: F, then: G) -> ModelWatcher<T>
    where
        T: Identifiable + 'static + Send + Sync,
        F: Fn(&T) -> bool + 'static + Send + Sync,
        G: Fn(&T) + 'static + Send + Sync,
    {
        let id = T::model_id();

        let (send, recv) = crossbeam::channel::unbounded();

        self.watcher.entry(id).or_default().push(send);

        ModelWatcher {
            filter: Box::new(filter),
            apply: Box::new(then),
            recv,
        }
    }

    pub fn get_id<T: for<'b> Deserialize<'b>>(&self, collection: &str, id: &str) -> T {
        unsafe { Store::get(collection, id, &*self.storage).unwrap() }
    }

    pub fn named(mut self) -> Self {
        self.named = true;
        self
    }

    pub fn with_cache(mut self, choice: bool) -> Self {
        self.cached = choice;
        self
    }

    /// Query `Vec<T>` which match `predicate`
    pub fn query<
        T: Identifiable + for<'b> Deserialize<'b> + Serialize + Send + Sync + 'static,
        F: Fn(&T) -> bool,
    >(
        &self,
        predicate: F,
    ) -> Vec<Model<T>> {
        query_t(predicate, self)
    }

    /// Delete an entry
    pub fn delete<
        X: Into<IdRef<T>>,
        T: 'static + Serialize + Identifiable + Send + Sync + Clone,
    >(
        &self,
        entry: X,
    ) {
        // delete from cache
        let entry: IdRef<T> = entry.into();
        let col = T::model_id();
        let id = entry.id.trim_start_matches(&format!("{col}::")).to_string();
        self.del_cache::<T>(id.clone());
        // delete from vfs
        self.storage
            .remove_file(&format!("/collection/{col}/{id}"))
            .unwrap();
    }

    /// Update every model in `entries` according to `u(_)`
    pub fn update<T: 'static + Serialize + Identifiable + Send + Sync + Clone, F: Fn(&mut T)>(
        &self,
        entries: &mut [Model<T>],
        u: F,
    ) {
        for e in entries {
            e.write(self, &u);
        }
    }

    fn ensure_init(db: Self) -> Self {
        let _ = db.storage.create_dir("/collection");
        let _ = db.storage.create_dir("/files");
        db
    }

    /// Create a new fs backed database at `path`
    pub fn filesystem(path: &str) -> Self {
        Self::ensure_init(Self {
            storage: Arc::new(vfs::PhysicalFS::new(path)),
            named: false,
            cached: true,
            records: Arc::new(DashMap::new()),
            watcher: Arc::new(DashMap::new()),
        })
    }

    /// List all collections
    pub fn list(&self) -> Vec<String> {
        Store::list(&*self.storage)
    }

    /// List all entries in `collection`
    pub fn list_entries(&self, collection: &str) -> Vec<String> {
        Store::get_ids(collection, &*self.storage)
    }

    pub fn try_cached<T: Identifiable + for<'b> Deserialize<'b> + 'static>(
        &self,
        col: String,
        id: String,
    ) -> Option<Model<T>> {
        // let id = id.trim_start_matches(&format!("{}::", T::model_id()));
        let entry = self.records.get(&col)?;
        let entry = entry.get(&id)?;
        let model: &Model<T> = entry.value().as_any().downcast_ref()?;
        Some(model.clone())
    }

    pub fn save_cache<T: Serialize + Identifiable + Send + Sync + 'static>(
        &self,
        id: String,
        model: Model<T>,
    ) {
        self.records
            .entry(T::model_id())
            .or_default()
            .insert(id, Box::new(model.clone()));
    }

    pub fn del_cache<T: Serialize + Identifiable + Send + Sync + 'static>(&self, id: String) {
        self.records.entry(T::model_id()).or_default().remove(&id);
    }

    /// Get a model `T` from `id`
    pub fn get<
        T: Identifiable + Serialize + Send + Sync + for<'b> Deserialize<'b> + 'static,
        I: Into<IdRef<T>>,
    >(
        &self,
        id: I,
    ) -> Option<Model<T>> {
        let id: IdRef<T> = id.into();
        let col = parse_collection(&id.id).unwrap_or(T::model_id());
        log::trace!("Getting {id} as {col}");
        if self.cached {
            if let Some(cached) = self.try_cached(col, id.to_string()) {
                log::trace!("Returning {id} from cache");
                return Some(cached);
            }
        }

        log::trace!("Trying to load {id}");
        let col = parse_collection(&id.to_string());
        let col = col.unwrap_or(T::model_id());
        let m: T = unsafe {
            Store::get(
                &col,
                &id.id.trim_start_matches(&format!("{}::", col)),
                &*self.storage,
            )?
        };
        let model = Model {
            inner: Arc::new(ArcSwap::new(Arc::new(m))),
        };

        if self.cached {
            self.save_cache(id.to_string(), model.clone());
            log::trace!("Returning {id} and saved to cache");
        }

        Some(model)
    }

    /// Save a raw model into the database
    pub unsafe fn save_raw<T: Serialize + Identifiable>(&self, data: &T, collection: &str) {
        Store::save(collection, data, self.named, &*self.storage);
    }

    fn save_model<T: Serialize + Identifiable + Send + Sync + 'static>(&self, data: &Model<T>) {
        if let Some(watchers) = self.watcher.get(&T::model_id()) {
            log::trace!("Notifying watchers");
            for watch in watchers.iter() {
                watch.send(Box::new(data.clone())).unwrap();
            }
        }

        let data = data.read();
        let data = (*data).deref();

        unsafe {
            Store::save(&T::model_id(), data, self.named, &*self.storage);
        }
    }

    /// Save a model `T` into the database
    pub fn save<'a, T: Serialize + Identifiable + Send + Sync + 'static>(
        &'a self,
        data: T,
    ) -> Model<T> {
        let model = Model {
            inner: Arc::new(ArcSwap::new(Arc::new(data))),
        };
        self.save_model(&model);
        let id = model.full_id().to_string();

        if self.cached {
            log::trace!("Saving {id} to cache");
            self.save_cache(id, model.clone());
        }

        model
    }
}

// TODO : cache eviction based on ref counts

pub struct Model<T> {
    inner: Arc<ArcSwap<T>>,
}

impl Model<File> {
    pub fn read_file(&self, db: &Database) -> Vec<u8> {
        self.read().read(db)
    }
}

impl<T: Identifiable> Model<T> {
    pub fn from_raw(raw: T) -> Model<T> {
        Model {
            inner: Arc::new(ArcSwap::new(Arc::new(raw))),
        }
    }

    pub fn full_id(&self) -> String {
        self.read().full_id()
    }

    pub fn reference(&self) -> IdRef<T> {
        self.read().reference()
    }

    pub fn read(&self) -> arc_swap::Guard<Arc<T>> {
        self.inner.load()
    }
}

impl<T: 'static + Serialize + Identifiable + Send + Sync + Clone> Model<T> {
    pub fn write<F: Fn(&mut T)>(&mut self, db: &Database, u: F) {
        let mut me = (**self.inner.load()).clone();
        u(&mut me);
        self.inner.swap(Arc::new(me));
        db.save_model(self);
    }

    /// Update the model inline without writing to `Database`
    pub fn write_raw_inline<F: Fn(&mut T)>(&mut self, u: F) {
        let mut me = (**self.inner.load()).clone();
        u(&mut me);
        drop(me);
    }
}

impl<T: 'static + Serialize + Identifiable + Send + Sync> ModelObj for Model<T> {
    fn as_any(&self) -> &dyn std::any::Any {
        self
    }
}

impl<T> Clone for Model<T> {
    fn clone(&self) -> Self {
        Self {
            inner: self.inner.clone(),
        }
    }
}

pub fn parse_collection(id: &str) -> Option<String> {
    if id.starts_with("[") {
        return None;
    }

    let splitted: Vec<_> = id.split("[").collect();
    if splitted.len() > 1 {
        let cols = splitted.first().unwrap();
        let col = cols.trim_end_matches("::");
        return Some(col.to_string());
    }

    let mut splitted = id.split("::").collect::<Vec<_>>();
    if splitted.len() == 1 {
        None
    } else {
        splitted.pop().unwrap();
        Some(splitted.join("::"))
    }
}
