use std::{io::Write, process::Stdio};

use owl_macro::model;
use sha2::{Digest, Sha256};

pub use crate as owl;
use crate::{
    db::Database,
    prelude::{Id, Saveable},
};

fn sha256(input: &[u8]) -> String {
    let mut hasher = Sha256::new();
    hasher.update(input);
    let result = hasher.finalize();
    format!("{:x}", result)
}

/// A generic file
#[derive(Debug)]
#[model]
pub struct File {
    id: Id,
    pub name: Option<String>,
    pub size: usize,
    pub mime: String,
}

pub fn get_mime_type(data: &[u8]) -> String {
    let mut child = std::process::Command::new("file")
        .arg("--brief")
        .arg("--mime-type")
        .arg("-") // Read from stdin
        .stdin(Stdio::piped())
        .stdout(Stdio::piped())
        .spawn()
        .unwrap();

    {
        let stdin = child.stdin.as_mut().ok_or("Failed to open stdin").unwrap();
        stdin.write_all(data).unwrap();
    }

    let output = child.wait_with_output().unwrap();

    if !output.status.success() {
        log::error!("Error getting MIME");
    }

    let mime_type = String::from_utf8(output.stdout).unwrap().trim().to_string();
    mime_type
}

impl File {
    pub fn new(data: Vec<u8>, name: Option<String>, db: &Database) -> Self {
        let hash = sha256(&data);

        let f = Self {
            id: Id::String(hash.clone()),
            size: data.len(),
            mime: get_mime_type(&data),
            name,
        };

        let first: String = hash.chars().take(2).collect();
        let second: String = hash.chars().skip(2).take(2).collect();

        let _ = db.storage.create_dir(&format!("/files/{first}"));
        let _ = db.storage.create_dir(&format!("/files/{first}/{second}"));
        db.storage
            .create_file(&format!("/files/{first}/{second}/{hash}"))
            .unwrap()
            .write(&data)
            .unwrap();

        unsafe {
            f.save(false, &*db.storage);
        }

        f
    }

    pub fn read(&self, db: &Database) -> Vec<u8> {
        let hash = &self.id.to_string();
        let first: String = hash.chars().take(2).collect();
        let second: String = hash.chars().skip(2).take(2).collect();

        let mut buf = Vec::with_capacity(self.size as usize);

        db.storage
            .open_file(&format!("/files/{first}/{second}/{hash}"))
            .unwrap()
            .read_to_end(&mut buf)
            .unwrap();

        buf
    }
}
