use rayon::iter::{IntoParallelRefIterator, ParallelIterator};
use serde::{Deserialize, Deserializer, Serialize};
use std::fmt::{Debug, Display};

use crate::Identifiable;

use super::{
    id::Id,
    store::{Saveable, Store},
    Database, Model,
};

#[allow(non_snake_case)]
pub mod RelationKind {
    /// Unidirectional relation kind
    ///
    /// The relation is the same for each node.
    pub struct Unidirectional;

    /// Unidirectional relation kind
    ///
    /// Relations can be directed.
    pub struct Bidirectional;
}

/// A relation between `T` and `S` with kind `U`.
/// Additionally with `weight` and `meta` values.
pub struct Relation<T, S, R, U> {
    pub name: String,
    pub top: std::marker::PhantomData<T>,
    pub sub: std::marker::PhantomData<S>,
    pub uni: std::marker::PhantomData<U>,
    pub references: std::marker::PhantomData<R>,
    pub weight: Option<f64>,
    pub meta: Option<serde_json::Value>,
}

impl<T, S, R, U> Relation<T, S, R, U> {
    /// Create a new relation with `name`
    pub fn new(name: &str) -> Self {
        Self {
            name: name.to_string(),
            weight: None,
            meta: None,
            references: std::marker::PhantomData,
            top: std::marker::PhantomData,
            sub: std::marker::PhantomData,
            uni: std::marker::PhantomData,
        }
    }
}

#[derive(PartialEq)]
/// A typed reference type
pub struct IdRef<T> {
    pub id: String,
    model: std::marker::PhantomData<T>,
}

impl<T> Debug for IdRef<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("IdRef").field(&self.id).finish()
    }
}

impl<T> Clone for IdRef<T> {
    fn clone(&self) -> Self {
        Self {
            id: self.id.clone(),
            model: self.model.clone(),
        }
    }
}

unsafe impl<T> Send for IdRef<T> {}
unsafe impl<T> Sync for IdRef<T> {}

impl<T> Display for IdRef<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(&self.id)
    }
}

impl<T: for<'a> Deserialize<'a>> From<String> for IdRef<T> {
    fn from(value: String) -> Self {
        IdRef::new(value)
    }
}

impl<T: for<'a> Deserialize<'a>> From<&str> for IdRef<T> {
    fn from(value: &str) -> Self {
        IdRef::new(value.to_string())
    }
}

impl<T: for<'a> Deserialize<'a>> From<Id> for IdRef<T> {
    fn from(value: Id) -> Self {
        IdRef::new(value.to_string())
    }
}

impl<T: for<'a> Deserialize<'a>, X: for<'a> Deserialize<'a>> From<&IdRef<X>> for IdRef<T> {
    fn from(value: &IdRef<X>) -> Self {
        IdRef::new(value.id.clone())
    }
}

impl<T: Identifiable + for<'a> Deserialize<'a>> From<&Model<T>> for IdRef<T> {
    fn from(value: &Model<T>) -> Self {
        IdRef::new(value.full_id())
    }
}

impl<T: Identifiable + for<'a> Deserialize<'a>> From<&T> for IdRef<T> {
    fn from(value: &T) -> Self {
        IdRef::new(value.full_id())
    }
}

impl<'de, T> Deserialize<'de> for IdRef<T> {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let s = String::deserialize(deserializer)?;
        Ok(IdRef {
            id: s,
            model: std::marker::PhantomData,
        })
    }
}

impl<T> Serialize for IdRef<T> {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        self.id.serialize(serializer)
    }
}

impl<T: for<'a> Deserialize<'a> + Send + Sync + Identifiable + Serialize + 'static> IdRef<T> {
    pub fn fetch_raw(&self, db: &Database) -> Model<T> {
        let id = self.id.clone();
        db.get::<T, _>(id).unwrap()
    }

    pub fn try_fetch(&self, db: &Database) -> Option<Model<T>> {
        // todo : asset type

        Some(self.fetch_raw(db))
    }
}

impl<T: for<'a> Deserialize<'a>> IdRef<T> {
    pub fn new(id: String) -> Self {
        Self {
            id,
            model: std::marker::PhantomData,
        }
    }

    pub unsafe fn fetch_raw_vfs(&self, vfs: &dyn vfs::FileSystem) -> T {
        Store::get_id(&self.id, vfs).unwrap()
    }

    pub unsafe fn fetch_as<O: for<'a> Deserialize<'a>>(&self, vfs: &dyn vfs::FileSystem) -> O {
        Store::get_id(&self.id, vfs).unwrap()
    }
}

impl<T: Identifiable + Serialize + Send + Sync + 'static + for<'a> Deserialize<'a>> IdRef<T> {
    /// Get the object this reference holds
    pub unsafe fn get_raw(&self, vfs: &dyn vfs::FileSystem) -> Option<T> {
        let split: Vec<&str> = self.id.split("::").collect();
        if split.len() > 1 {
            Store::get_id(&self.id, vfs)
        } else {
            Store::get(&T::model_id(), &self.id, vfs)
        }
    }

    pub fn dereference(&self, db: &Database) -> Model<T> {
        self.try_dereference(db).unwrap()
    }

    pub fn try_dereference(&self, db: &Database) -> Option<Model<T>> {
        db.get(self.id.clone())
    }
}

impl<
        T: Identifiable + Saveable + Send + Sync + 'static,
        S: Identifiable + Saveable + Send + Sync + 'static,
        R: RelationRef,
        U,
    > Relation<T, S, R, U>
{
    pub unsafe fn top(reference: &Model<R>, db: &Database) -> Model<T> {
        db.get(reference.read().top().id.clone()).unwrap()
    }

    pub unsafe fn sub(reference: &Model<R>, db: &Database) -> Model<S> {
        db.get(reference.read().sub().id.clone()).unwrap()
    }

    /// Parse a relation ref id into the respective pointer structure.
    ///
    /// The ref id looks like: `[top_id->sub_id]`
    pub fn parse_ref_id(id: &str) -> (String, String) {
        // Expected format: "[top_id->sub_id]"
        if let Some(stripped) = id.strip_prefix('[').and_then(|s| s.strip_suffix(']')) {
            if let Some((top, sub)) = stripped.split_once("->") {
                return (top.to_string(), sub.to_string());
            }
        }
        panic!("Invalid ref_id format: {}", id);
    }

    /// Return a relation ref id for `top` and `sub`
    pub fn ref_id(top: &Model<T>, sub: &Model<S>) -> String {
        format!("[{}->{}]", top.reference(), sub.reference())
    }

    /// Return a collection id for this relation
    pub fn collection_id(&self) -> String {
        format!("{}-refs", self.name)
    }
}

impl<
        T: Identifiable + Saveable + Send + Sync + 'static,
        S: Identifiable + Saveable + Send + Sync + 'static,
        R: RelationRef,
    > Relation<T, S, R, RelationKind::Unidirectional>
{
    /// Get a `RelationReference` if a relation exists between `top` and `sub`
    pub fn get(&self, top: &Model<T>, sub: &Model<S>, db: &Database) -> Option<Model<R>> {
        let entry = db.get(Relation::<T, S, R, RelationKind::Unidirectional>::ref_id(
            top, sub,
        ));
        let entry_rev = db.get(Relation::<S, T, R, RelationKind::Unidirectional>::ref_id(
            sub, top,
        ));

        if let Some(entry) = entry {
            return Some(entry);
        }
        if let Some(entry) = entry_rev {
            return Some(entry);
        }

        None
    }

    /// Get all `RelationReference`s with `node`
    pub fn get_all_with_t<X: Into<IdRef<T>>>(
        &self,
        node: X,
        vfs: &dyn vfs::FileSystem,
    ) -> Vec<IdRef<R>> {
        let ids = Store::get_ids(&self.collection_id(), vfs);
        let node: IdRef<T> = node.into();
        let top_id = node.id;
        ids.into_iter()
            .filter_map(|x| {
                let (t_id, s_id) =
                    Relation::<T, S, R, RelationKind::Unidirectional>::parse_ref_id(&x);
                if top_id == t_id || top_id == s_id {
                    return Some(IdRef::new(x.clone()));
                } else {
                    return None;
                }
            })
            .collect()
    }

    /// Get all `RelationReference`s with `node`
    pub fn get_all_with_s<X: Into<IdRef<S>>>(
        &self,
        node: X,
        vfs: &dyn vfs::FileSystem,
    ) -> Vec<IdRef<R>> {
        let ids = Store::get_ids(&self.collection_id(), vfs);
        let node: IdRef<S> = node.into();
        let top_id = node.id;
        ids.into_iter()
            .filter_map(|x| {
                let (t_id, s_id) =
                    Relation::<T, S, R, RelationKind::Unidirectional>::parse_ref_id(&x);
                if top_id == t_id || top_id == s_id {
                    return Some(IdRef::new(x.clone()));
                } else {
                    return None;
                }
            })
            .collect()
    }

    /// Check if a relation exists between `top` and `sub`
    pub fn exists<V: vfs::FileSystem>(&self, top: &Model<T>, sub: &Model<S>, vfs: &V) -> bool {
        let entry: Option<RelationReference> = unsafe {
            Store::get_opt(
                &self.collection_id(),
                &Relation::<T, S, R, RelationKind::Unidirectional>::ref_id(&top, &sub),
                vfs,
            )
        };
        let entry_rev: Option<RelationReference> = unsafe {
            Store::get_opt(
                &self.collection_id(),
                &Relation::<S, T, R, RelationKind::Unidirectional>::ref_id(&sub, &top),
                vfs,
            )
        };
        if entry.is_some() || entry_rev.is_some() {
            return true;
        }

        false
    }

    /// Add a relation for `top` and `sub` with optional `weight` and `meta` values
    pub fn add(
        &self,
        top: &Model<T>,
        sub: &Model<S>,
        weight: Option<f64>,
        meta: Option<serde_json::Value>,
        db: &Database,
    ) {
        let mut reference = self.get(&top, &sub, db).unwrap_or_else(|| {
            Model::from_raw(R::from_ref(RelationReference {
                top: IdRef::new(top.full_id()),
                sub: IdRef::new(sub.full_id()),
                uni: true,
                weight: weight,
                meta: None,
            }))
        });
        reference.write(db, |r| {
            r.update_meta(weight, meta.clone());
        });
    }
}

impl<
        T: Identifiable + Saveable + Send + Sync + 'static,
        S: Identifiable + Saveable + Send + Sync + 'static,
        R: RelationRef,
    > Relation<T, S, R, RelationKind::Bidirectional>
{
    /// Get a `RelationReference` if a relation exists between `top` and `sub`
    pub unsafe fn get(&self, top: &Model<T>, sub: &Model<S>, db: &Database) -> Option<Model<R>> {
        let entry = db.get(Relation::<T, S, R, RelationKind::Bidirectional>::ref_id(
            top, sub,
        ));
        if let Some(entry) = entry {
            return Some(entry);
        }

        None
    }

    /// Get all `RelationReference`s with `node` as top
    pub fn get_all_top(&self, top: &Model<T>, vfs: &dyn vfs::FileSystem) -> Vec<IdRef<R>> {
        let ids = Store::get_ids(&self.collection_id(), vfs);
        let top_id = top.full_id();
        ids.into_iter()
            .filter_map(|x| {
                let (t_id, _) = Relation::<T, S, R, RelationKind::Unidirectional>::parse_ref_id(&x);
                if top_id == t_id {
                    return Some(IdRef::new(format!("{}::{x}", R::model_id())));
                } else {
                    return None;
                }
            })
            .collect()
    }

    /// Get all `RelationReference`s with `node` as sub
    pub fn get_all_sub(&self, sub: &Model<T>, vfs: &dyn vfs::FileSystem) -> Vec<IdRef<R>> {
        let ids = Store::get_ids(&self.collection_id(), vfs);
        let sub_id = sub.full_id();
        ids.into_iter()
            .filter_map(|x| {
                let (_, s_id) = Relation::<T, S, R, RelationKind::Unidirectional>::parse_ref_id(&x);
                if sub_id == s_id {
                    return Some(IdRef::new(x));
                } else {
                    return None;
                }
            })
            .collect()
    }

    /// Check if a relation exists between `top` and `sub`
    pub fn exists(&self, top: &Model<T>, sub: &Model<S>, vfs: &dyn vfs::FileSystem) -> bool {
        let entry: Option<RelationReference> = unsafe {
            Store::get_opt(
                &self.collection_id(),
                &Relation::<T, S, R, RelationKind::Bidirectional>::ref_id(&top, &sub),
                vfs,
            )
        };
        if entry.is_some() {
            return true;
        }

        false
    }

    /// Add a relation for `top` and `sub` with optional `weight` and `meta` values
    pub fn add(
        &self,
        top: &Model<T>,
        sub: &Model<S>,
        weight: Option<f64>,
        meta: Option<serde_json::Value>,
        db: &Database,
    ) {
        let mut reference = unsafe {
            self.get(&top, &sub, db).unwrap_or_else(|| {
                Model::from_raw(R::from_ref(RelationReference {
                    top: IdRef::new(top.full_id()),
                    sub: IdRef::new(sub.full_id()),
                    uni: true,
                    weight: weight,
                    meta: None,
                }))
            })
        };
        reference.write(db, |reference: &mut _| {
            reference.update_meta(weight, meta.clone());
        });
    }
}

pub trait RelationRef:
    Identifiable + for<'a> Deserialize<'a> + Serialize + Send + Sync + 'static + Clone
{
    fn top(&self) -> IdRef<serde_json::Value>;
    fn sub(&self) -> IdRef<serde_json::Value>;
    fn from_ref(r: RelationReference) -> Self;
    fn update_meta(&mut self, weight: Option<f64>, meta: Option<serde_json::Value>);
}

#[derive(Deserialize, Serialize, Debug, Clone)]
pub struct RelationReference {
    pub top: IdRef<serde_json::Value>,
    pub sub: IdRef<serde_json::Value>,
    pub uni: bool,
    pub weight: Option<f64>,
    pub meta: Option<serde_json::Value>,
}

impl RelationReference {
    pub fn alter_meta<F: Fn(&mut X), X: for<'a> Deserialize<'a> + Default + Serialize>(
        &mut self,
        u: F,
    ) {
        let mut meta =
            serde_json::from_value(self.meta.clone().unwrap()).unwrap_or_else(|_| X::default());
        u(&mut meta);
        self.meta = Some(serde_json::to_value(meta).unwrap());
    }

    pub fn id(&self) -> String {
        format!("[{}->{}]", self.top.id, self.sub.id)
    }

    /*
    pub unsafe fn save(&self, vfs: &dyn vfs::FileSystem) {
        Store::save_raw(&self.model_id_real(), &self.id(), self, false, vfs);
    }
    */
}

/// Dereference a `Vec<_>` with `IdRef<T>`s into their objects
pub fn dereference<
    T: for<'a> Deserialize<'a> + Serialize + Identifiable + Send + Sync + 'static,
>(
    v: &[IdRef<T>],
    db: &Database,
) -> Vec<Model<T>> {
    v.into_iter().map(|x| x.fetch_raw(db)).collect()
}

pub unsafe fn dereference_raw<T: for<'a> Deserialize<'a>>(v: &[IdRef<T>], db: &Database) -> Vec<T> {
    v.into_iter()
        .map(|x| unsafe { x.fetch_raw_vfs(&*db.storage) })
        .collect()
}

// TODO : generally clean up traversal loops
pub fn clean_graph_traversal<T>(id: &str, route: &[IdRef<T>]) -> Vec<IdRef<T>> {
    let mut index = 0;

    for (i, node) in route.iter().enumerate() {
        let node = node.id.to_string();
        if node == id {
            index = i + 1;
        }
    }

    route[index..].to_vec()
}

/// Find a path from `from` to `to` with max `reach` steps to get there.
///
/// `fetch_relation` is a function for traversing the graph ( `Fn(IdRef<T>, &Database) -> Vec<IdRef<T>>` )
pub fn find_path<F, T: Debug>(
    from: IdRef<T>,
    to: IdRef<T>,
    reach: u32,
    fetch_relation: F,
    db: &Database,
) -> Vec<IdRef<T>>
where
    F: Fn(&IdRef<T>, &Database) -> Vec<IdRef<T>> + Sync + Send + Clone,
{
    log::info!("path from {from} r {reach}");
    let my_friends = fetch_relation(&from, db);
    log::trace!("have children {my_friends:?}");

    if reach == 0 {
        log::trace!("no reach :(");
        return vec![];
    }

    my_friends
        .par_iter()
        .filter_map(|friend| {
            if friend.id == to.id {
                log::trace!("found you {friend}");
                return Some(vec![friend.clone()]);
            }

            log::trace!("nooo... looking deeper");
            let p = find_path(
                friend.clone(),
                to.clone(),
                reach - 1,
                fetch_relation.clone(),
                db,
            );
            if !p.is_empty() {
                let mut ret = vec![friend.clone()];
                ret.extend(p.iter().cloned());
                return Some(ret);
            }

            None
        })
        .find_any(|_| true) // find any successful path in parallel
        .unwrap_or_else(|| {
            log::trace!("END SEARCH");
            vec![]
        })
}

/// Get the other which isnt oneself from unidirectional `RelationReference`s
pub fn get_other<
    X: Identifiable + Send + Sync + Serialize + for<'a> Deserialize<'a>,
    T: Into<IdRef<X>>,
    R: RelationRef,
>(
    id: T,
    refs: Vec<IdRef<R>>,
    db: &Database,
) -> Vec<IdRef<serde_json::Value>> {
    let mut other = Vec::new();
    let id: IdRef<X> = id.into();
    let id = id.id;

    for r in refs {
        if let Some(r) = r.try_dereference(&db) {
            if r.read().top().id != id {
                other.push(r.read().top());
            } else {
                other.push(r.read().sub());
            }
        } else {
            log::error!("Could not get Model<{}> from {}", R::model_id(), r.id);
        }
    }

    other
}
