use serde::{Deserialize, Serialize};

use crate::Identifiable;

fn read_to_vec(path: &str, vfs: &dyn vfs::FileSystem) -> Option<Vec<u8>> {
    let mut f = vfs.open_file(path).ok()?;
    let mut content = Vec::new();
    f.read_to_end(&mut content).ok()?;
    Some(content)
}

pub fn write(path: &str, data: &[u8], vfs: &dyn vfs::FileSystem) {
    let mut f = vfs.create_file(path).unwrap();
    f.write_all(&data).unwrap();
}

pub struct Store {}

impl Store {
    pub unsafe fn get<T: for<'a> Deserialize<'a>>(
        collection: &str,
        id: &str,
        vfs: &dyn vfs::FileSystem,
    ) -> Option<T> {
        let path = std::path::Path::new(&format!("/collection/{collection}"))
            .join(id)
            .display()
            .to_string();
        log::info!("IO READ {path}");
        let content = read_to_vec(&path, vfs)?;
        rmp_serde::from_slice(&content).ok()
    }

    pub unsafe fn get_opt<T: for<'a> Deserialize<'a>>(
        collection: &str,
        id: &str,
        vfs: &dyn vfs::FileSystem,
    ) -> Option<T> {
        let path = std::path::Path::new(&format!("/collection/{collection}"))
            .join(id)
            .display()
            .to_string();
        rmp_serde::from_slice(&read_to_vec(&path, vfs)?).unwrap()
    }

    pub unsafe fn save_raw<T: Serialize>(
        collection: &str,
        id: &str,
        data: &T,
        named: bool,
        vfs: &dyn vfs::FileSystem,
    ) {
        let c = if named {
            rmp_serde::to_vec_named(data).unwrap()
        } else {
            rmp_serde::to_vec(data).unwrap()
        };

        let path = std::path::Path::new(&format!("/collection/{collection}"))
            .join(id)
            .display()
            .to_string();
        let _ = vfs.create_dir(&format!("/collection/{collection}"));
        log::info!("IO WRITE {path}");
        write(&path, &c, vfs);
    }

    pub unsafe fn save<T: Serialize + Identifiable>(
        collection: &str,
        data: &T,
        named: bool,
        vfs: &dyn vfs::FileSystem,
    ) {
        Self::save_raw(collection, &data.id().to_string(), &data, named, vfs);
    }

    pub unsafe fn get_id<T: for<'a> Deserialize<'a>>(
        id: &str,
        vfs: &dyn vfs::FileSystem,
    ) -> Option<T> {
        let split_d: Vec<_> = id.split("[").collect();
        let id = split_d.first()?;
        let mut split: Vec<_> = id.split("::").collect();
        let sid: &str = split.pop()?;
        let collection = split.join("::");
        let id = format!(
            "{}{sid}{}",
            if split_d.len() > 1 { "[" } else { "" },
            split_d
                .iter()
                .skip(1)
                .map(|x| x.to_string())
                .collect::<Vec<_>>()
                .join("[")
        );
        Store::get(&collection, &id, vfs)
    }

    pub fn get_ids(collection: &str, vfs: &dyn vfs::FileSystem) -> Vec<String> {
        vfs.read_dir(&format!("/collection/{collection}"))
            .unwrap()
            .collect()
    }

    pub fn list(vfs: &dyn vfs::FileSystem) -> Vec<String> {
        vfs.read_dir("/collection").unwrap().collect()
    }
}

pub trait Saveable: Identifiable + for<'a> Deserialize<'a> + Serialize {
    unsafe fn save(&self, named: bool, vfs: &dyn vfs::FileSystem) {
        Store::save(&Self::model_id(), self, named, vfs);
    }

    unsafe fn get(id: &str, vfs: &dyn vfs::FileSystem) -> Option<Self> {
        Store::get(&Self::model_id(), id, vfs)
    }
}
