//mod model;

use db::id::Id;

pub mod db;
use once_cell::sync::OnceCell;
pub use owl_macro::model;
pub use owl_macro::relation;
pub use serde::{Deserialize, Serialize};

pub static DB: OnceCell<db::Database> = OnceCell::new();

/// Set a `Database` as global.
///
/// All global macro actions are then done against that `Database`.
#[macro_export]
macro_rules! set_global_db {
    ($db:ident) => {
        assert!($crate::DB.set($db).is_ok());
    };
}

#[macro_export]
macro_rules! dereference {
    ($id:ident) => {
        $id.dereference(&$crate::DB.get().unwrap())
    };
    ($id:expr) => {
        $id.dereference(&$crate::DB.get().unwrap())
    };
}

#[macro_export]
macro_rules! save {
    ($model:ident) => {
        $crate::DB.get().unwrap().save($model)
    };
    ($model:expr) => {
        $crate::DB.get().unwrap().save($model)
    };
}

#[macro_export]
macro_rules! get {
    ($id:ident) => {
        $crate::DB.get().unwrap().get($id)
    };
    ($id:expr) => {
        $crate::DB.get().unwrap().get($id)
    };
}

#[macro_export]
macro_rules! update {
    ($ids:ident, $u:expr) => {
        $crate::DB.get().unwrap().update($ids, $u)
    };
    ($ids:expr, $u:expr) => {
        $crate::DB.get().unwrap().update($ids, $u)
    };
}

#[macro_export]
macro_rules! query {
    ($id:ident) => {
        $crate::DB.get().unwrap().query($id)
    };
    ($id:expr) => {
        $crate::DB.get().unwrap().query($id)
    };
}

pub mod prelude {
    pub use super::db::field::Historic;
    pub use super::db::id::Id;
    pub use super::db::relation::{dereference, IdRef, Relation, RelationKind};
    pub use super::db::store::{Saveable, Store};
    pub use super::db::Database;
    pub use super::db::Model;
    pub use super::Identifiable;
    pub use super::{model, relation};
}

// TODO : wire format for endpoints
// TODO : auth + traceability
// TODO : deletion?
// TODO : make relations able to be inactive/active

pub trait Identifiable: Sized + for<'a> Deserialize<'a> {
    fn id(&self) -> Id;

    /// Get a typed `IdRef<T>` for this ID
    fn reference(&self) -> crate::db::relation::IdRef<Self> {
        crate::db::relation::IdRef::new(self.full_id())
    }
    fn model_id() -> String;
    fn full_id(&self) -> String {
        format!("{}::{}", Self::model_id(), self.id())
    }
}

impl Identifiable for serde_json::Value {
    fn id(&self) -> Id {
        serde_json::from_value(self.as_object().unwrap().get("id").unwrap().clone()).unwrap()
    }

    fn model_id() -> String {
        "json".to_string()
    }
}
